import json
import logging
import sys
from collections import defaultdict
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

from datamodel_code_generator.config import JSONSchemaParserConfig
from datamodel_code_generator.model import pydantic_v2
from datamodel_code_generator.parser.base import Result
from datamodel_code_generator.parser.jsonschema import JsonSchemaParser
from pyang.context import Context
from pyang.statements import ModSubmodStatement, Statement
from pydantic import BaseModel
from typing_extensions import Self

from ..models import ModelRoot, Node
from ..utility import restconf_patch_request
from . import (
    YANGSourcesTracker,
    function_content_to_source_code,
    function_to_source_code,
)

logger = logging.getLogger("pydantify")


# Helper function
def dynamically_serialized_helper_function():  # pragma: no cover
    if __name__ == "__main__":
        # Demonstration purposes only. Not included in actual output.
        # To run: pdm run python out/out.py
        from pathlib import Path

        with open(Path(__file__).parent.joinpath("sample_data.json")) as fd:
            import json

            input = json.load(fd)
            print(f"{input=}")
            model = Model(**input)
            print("Instantiation successful!")

            output = model.model_dump_json(exclude_defaults=True, by_alias=True)
            print(f"{output=}")

            assert json.loads(output) == input
            print("Serialization successful!")


def model_init_code():  # pragma: no cover
    if __name__ == "__main__":
        model = Model(
            # <Initialize model here>
        )

        restconf_payload = model.model_dump_json(
            exclude_defaults=True, by_alias=True, indent=2
        )

        print(f"Generated output: {restconf_payload}")


class ModelGenerator:
    """In charge of generating the output model"""

    include_verification_code: bool = False
    input_dir: Path
    output_dir: Path
    standalone: bool = False
    trim_path: Optional[str] = None
    json_schema_output: bool

    @classmethod
    def generate(
        cls: Type[Self],
        ctx: Context,
        modules: List[ModSubmodStatement],
        fd: TextIOWrapper,
    ):
        """Generate and write output model to a given file descriptor."""
        # Generate actual model
        cls.__generate(modules, fd)
        fd.write("\n\n")

        # Add initialization helper-code if Pydantic models generated
        if cls.json_schema_output is False:
            cls.__generate_helper_code(fd)

    @classmethod
    def __generate(
        cls: Type[Self], modules: List[ModSubmodStatement], fd: TextIOWrapper
    ):
        """Generates and yields"""

        for module in modules:
            if cls.trim_path is not None:
                split_path = cls.split_path(cls.trim_path)
                module = cls.trim(module, split_path)
            if module is None:
                logger.error("Invalid module. Exiting.")
                sys.exit(0)
            mod = ModelRoot(module)
            pydantic_model = mod.to_pydantic_model()
            if pydantic_model is None:
                continue
            # Skip models that only have 'prefix' and/or 'namespace' fields
            model_fields = set(pydantic_model.model_fields.keys())
            if model_fields <= {"prefix", "namespace"}:
                continue
            schema = cls.custom_dump(pydantic_model)
            result = (
                json.dumps(schema, indent=2)
                if cls.json_schema_output is True
                else cls.__generate_pydantic(json.dumps(schema))
            )
            # Keep mypy happy
            if isinstance(result, str):
                fd.write(result)
            else:
                logger.warning(
                    f"Expected string but got {type(result)} whilst parsing JSON"
                )
            pass

    @staticmethod
    def __generate_pydantic(json: str) -> str | dict[tuple[str, ...], Result]:
        """Generates pydantic models"""
        extra_template_data: defaultdict[str, dict[str, Any]] = defaultdict(dict)
        extra_template_data["#all#"]["config"] = {}
        extra_template_data["#all#"]["config"]["regex_engine"] = '"python-re"'
        parser = JsonSchemaParser(
            json,
            config=JSONSchemaParserConfig(
                data_model_type=pydantic_v2.BaseModel,
                data_model_root_type=pydantic_v2.RootModel,
                data_type_manager_type=pydantic_v2.DataTypeManager,
                data_model_field_type=pydantic_v2.DataModelField,
                snake_case_field=True,
                apply_default_values_for_required_fields=True,
                use_annotated=True,
                field_constraints=True,
                use_schema_description=True,
                use_field_description=True,
                aliases=Node.alias_mapping,
                reuse_model=False,  # Causes DCG to aggressively re-use "equivalent" classes, even if unrelated.
                allow_population_by_field_name=True,
                allow_extra_fields=False,
                collapse_root_models=True,
                extra_template_data=extra_template_data,
                field_extra_keys=set(["x-is-classvar"]),
            ),
        )
        return parser.parse()

    @classmethod
    def __generate_helper_code(cls: Type[Self], fd: TextIOWrapper) -> None:
        if cls.standalone:
            fd.write(function_to_source_code(restconf_patch_request))
            fd.write("\n\n")

        if cls.include_verification_code:
            fd.write(
                function_content_to_source_code(dynamically_serialized_helper_function)
            )
        else:
            fd.write(function_content_to_source_code(model_init_code))
            fd.write("\n")
            fd.write(
                "\n".join(
                    [
                        "    # Send config to network device:",
                        (
                            "    # from pydantify.utility import restconf_patch_request"
                            if not cls.standalone
                            else ""
                        ),
                        "    # restconf_patch_request(url='...', user_pw_auth=('usr', 'pw'), data=restconf_payload)",
                    ]
                )
            )
        YANGSourcesTracker.copy_yang_files(
            input_root=cls.input_dir, output_dir=cls.output_dir
        )

    @classmethod
    def split_path(cls: Type[Self], path: str) -> List[str]:
        return [p for p in path.split("/") if p != ""]

    @classmethod
    def trim(
        cls: Type[Self], statement: Statement, path: List[str]
    ) -> Statement | None:
        arg, path = path[0], path[1:]
        if arg == statement.arg:
            while len(path) > 0:
                for child in statement.i_children:
                    if child.arg == path[0]:
                        statement = child
                        path = path[1:]
                        break
                else:
                    logger.warn(
                        f'Path element "{path[0]}" not found in "{statement.arg}\n'
                        f"Available: [{', '.join(ch.arg for ch in statement.i_children)}]"
                    )
                    return None
            return statement
        return None

    @classmethod
    def custom_dump(cls: Type[Self], model: Type[BaseModel]) -> Dict[str, Any]:
        schema = model.model_json_schema(by_alias=True)
        if cls.json_schema_output:
            properties = schema.get("properties", {})
            properties.pop("namespace", None)
            properties.pop("prefix", None)
            for def_schema in schema["$defs"].values():
                def_properties = def_schema.get("properties", {})
                def_properties.pop("namespace", None)
                def_properties.pop("prefix", None)
        return schema



import logging
import os
import time
from io import TextIOWrapper
from typing import Dict

import psutil
from pyang.context import Context
from pyang.plugin import PyangPlugin, register_plugin
from pyang.statements import ModSubmodStatement

from pydantify.utility.model_generator import ModelGenerator

logger = logging.getLogger("pydantify")


def pyang_plugin_init():
    register_plugin(Pydantify())
    logger.debug("Plugin successfully registered.")


class Pydantify(PyangPlugin):
    def __init__(self):
        """Init plugin instance."""
        super().__init__(name="pydantify")

    def add_output_format(self, fmts: Dict[str, PyangPlugin]):
        """Register self as primary pydantic output generator."""
        fmts["pydantic"] = self
        self.multiple_modules = True
        self.handle_comments = True
        logger.debug("Plugin registered for pydantic format.")

    def emit(self, ctx: Context, modules: ModSubmodStatement, fd: TextIOWrapper):
        """Convert yang model."""
        start = psutil.Process(os.getpid()).create_time()
        logger.debug(f"Pyang completed parsing in {time.time() - start:.3f}s.")

        start = time.time()
        ModelGenerator.generate(ctx=ctx, modules=modules, fd=fd)
        logger.info(f"Output model generated in {time.time() - start:.3f}s.")



from lxml import etree
from pydantic import BaseModel
from typing import Any


def model_dump_xml_string(model: BaseModel, pretty_print: bool = False) -> str:
    element = model_dump_xml(model)
    return etree.tostring(element, encoding=str, pretty_print=pretty_print)


def model_dump_xml(model: BaseModel) -> etree.Element:
    dump = model.model_dump(exclude_defaults=True)
    return dict_to_elements(dump, model.namespace, model.prefix)[0]  # type: ignore  # ToDo: Find a better way


def dict_to_elements(
    model: dict[str, Any], namespace: str, prefix: str
) -> list[etree.Element]:
    result: list[etree.Element] = []
    for name, value in model.items():
        element = etree.Element(f"{{{namespace}}}{name}", nsmap={prefix: namespace})
        if isinstance(value, dict):
            for child in dict_to_elements(value, namespace, prefix):
                element.append(child)
        else:
            element.text = str(value)
        result.append(element)
    return result



def restconf_patch_request(url: str, user_pw_auth: tuple[str, str], data: str):
    r"""Sends a restconf "PATCH" request to a network device

    :param url: e.g. `'https://sandbox.cisco.com/restconf/data/Cisco-IOS-XE-native:native/interface'`
    :param user_pw_auth: tuple of username & password, e.g. `("user", "pw")`
    :param data: restconf payload as generated by `Model.json(...)`

    :return: HTTP Request response
    """
    import requests

    response = requests.patch(
        url=url,
        auth=user_pw_auth,
        headers={
            "Accept": "application/yang-data+json",
            "Content-Type": "application/yang-data+json",
        },
        data=data,
        verify=False,
    )

    return response



import logging
import os
import shutil
from pathlib import Path
from typing import Set, Type

from pyang.error import Position
from typing_extensions import Self

logger = logging.getLogger("pydantify")


class YANGSourcesTracker:
    __relevant_files: Set[str]

    @classmethod
    def track_from_pos(cls: Type[Self], pos: Position) -> None:
        path = str(Path(pos.ref).absolute())
        cls._relevant_files().add(path)

    # TODO: Is this necessary?
    # Pytest does not reset static class variable otherwise.
    @classmethod
    def _relevant_files(cls) -> Set[str]:
        if not hasattr(cls, "__relevant_files"):
            cls.__relevant_files = set()
        return cls.__relevant_files

    @classmethod
    def copy_yang_files(cls: Type[Self], input_root: Path, output_dir: Path) -> None:
        """Copy only the relevant YANG model files to the output directory."""
        for f in cls._relevant_files():
            out_path = output_dir
            if input_root is not None:
                delta = Path(f).parent.relative_to(input_root)
                out_path = output_dir.joinpath(delta)
                if not out_path.exists():
                    os.makedirs(out_path, exist_ok=True)
            out = shutil.copy2(f, out_path)
            logger.debug(f'Copied file "{f}" -> "{out}"')



#!/usr/bin/env python

import logging
import sys
from typing import List

logging.basicConfig(
    stream=sys.stdout,
    format="[%(levelname)s] %(pathname)s:%(lineno)d (%(funcName)s): %(message)s",
    level=logging.INFO,
)
logger = logging.getLogger("pydantify")


def main():
    from pyang.scripts.pyang_tool import run

    # Parse user-give settings
    sys.argv[1:] = parse_cli_arguments()

    run()


def parse_cli_arguments() -> List[str]:
    """Parses and handles incoming CLI arguments and returns all the ones destined for Pydantic."""

    # Import locally to not clutter scope of caller
    import logging
    import os
    from argparse import ArgumentParser
    from pathlib import Path

    from .models.base import Node
    from .utility.model_generator import ModelGenerator

    # Setup parser
    parser = ArgumentParser(
        prog="pydantify",
        description="Transform a YANG model to a serializable Pydantic model.",
        epilog="NOTE: All unknown arguments will be passed to Pyang as-is and without guarantees.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Enables debug output",
        default=False,
    )
    parser.add_argument(
        "-V",
        "--include-verification",
        action="store_true",
        dest="verify",
        help="Adds validation code, as well as the relevant YANG files, to the output model.",
        default=False,
    )
    parser.add_argument(
        "-S",
        "--standalone",
        action="store_true",
        dest="standalone",
        help="Generated output model has no dependency on Pydantify. All required code is copied into the output model.",
        default=False,
    )
    parser.add_argument(
        "-i",
        "--input-dir",
        "--path",
        dest="input_dir",
        help="The directory that contains the YANG input model. Defaults to the input file's folder.",
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        help='The directory that should be used to store the output model. Defaults to "$CWD/out".',
        default=f"{os.getcwd()}/out/",
    )
    parser.add_argument(
        "-f",
        "--output-file",
        dest="output_file",
        help='The name of the output file. Defaults to "out.py" for Pydantic output or "out.json" for JSON schema output.',
        default=None,
    )
    parser.add_argument(
        "input_file",
        action="store",
        help="The YANG file containing the entrypoint to the model to evaluate.",
    )
    parser.add_argument(
        "-t",
        "--trim-path",
        dest="trim_path",
        help="Get only the specified branch of the whole tree.",
        default=None,
    )
    parser.add_argument(
        "-j",
        "--json-schema",
        action="store_true",
        dest="json_schema_output",
        help="Output JSON schema instead of Pydantic models.",
        default=False,
    )
    parser.add_argument(
        "-d",
        "--data-type",
        action="store",
        dest="data_type",
        choices=["config", "state"],
        help="Limit output to config or state only. Default is config and state combined.",
        default=None,
    )
    parser.add_argument(
        "-n",
        "--strip-namespace",
        action="store_true",
        dest="strip_namespace",
        help="Strip the YANG namespace from the output model aliases.",
        default=False,
    )
    relay_args: List[str] = []

    # Parse
    args, unknown_args = parser.parse_known_args()

    # Apply known settings accordingly
    logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    ModelGenerator.include_verification_code = args.verify
    ModelGenerator.standalone = args.standalone
    ModelGenerator.json_schema_output = args.json_schema_output
    Node.data_type = args.data_type
    Node.strip_namespace = args.strip_namespace
    default_output_file = "out.json" if args.json_schema_output else "out.py"

    input_dir = (
        Path(args.input_file).absolute().parent
        if args.input_dir is None
        else Path(args.input_dir).absolute()
    )
    ModelGenerator.input_dir = input_dir
    relay_args.append(f"--path={input_dir}")

    ModelGenerator.trim_path = args.trim_path

    output_dir = Path(args.output_dir).absolute()
    ModelGenerator.output_dir = output_dir
    os.makedirs(output_dir, exist_ok=True)  # Create output directory if not exists
    if args.json_schema_output is False:
        with open(output_dir / "__init__.py", "a"):  # Create init file if not exists
            pass
    relay_args.append(
        f"--output={output_dir}/{args.output_file if args.output_file is not None else default_output_file}"
    )
    relay_args.append(f"--plugindir={Path(__file__).parent}/plugins")
    relay_args.append("--format=pydantic")

    # Order of arguments matters.
    return relay_args + unknown_args + [args.input_file]



from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Union, Optional

from pyang.statements import (
    ChoiceStatement,
    ContainerStatement,
    LeafLeaflistStatement,
    ListStatement,
    ModSubmodStatement,
    Statement,
    TypedefStatement,
)
from pydantic import create_model
from pydantic.fields import Field, FieldInfo

from . import BaseModel, GeneratedClass, Node, NodeFactory, RootModel, TypeResolver

if TYPE_CHECKING:
    __class__: type

logger = logging.getLogger("pydantify")


class TypeDefNode(Node):
    def __init__(self, stm: TypedefStatement) -> None:
        super().__init__(stm)

        self._output_model = GeneratedClass(
            class_name=self.name(),
            field_info=Field(
                self.default if self.default is not None or not self.mandatory else ...,
                alias=self.get_qualified_name(),
            ),  # type: ignore[arg-type]
            cls=self.to_pydantic_model(),
        )

    def get_base_class(self) -> type | Node | Enum:
        base = TypeResolver.resolve_statement(self.raw_statement)
        return base

    def name(self) -> str:
        return self.make_unique_name(suffix="Type")

    def to_pydantic_model(self) -> type[RootModel]:
        """Generates the output class representing this Typedef."""
        base_type = TypeResolver.resolve_statement(self.raw_statement)
        output_model: type[RootModel] = create_model(
            self.name(),
            __base__=(RootModel[base_type],),  # type: ignore[misc]
        )
        output_model.__doc__ = self.description or ""
        return output_model


@NodeFactory.register_statement_class(["leaf"])
class LeafNode(Node):
    def __init__(self, stm: LeafLeaflistStatement) -> None:
        logger.debug(f"Parsing {__class__}")
        super().__init__(stm)

        self._output_model = GeneratedClass(
            class_name=self.name(),
            field_info=Field(
                self.default if self.default is not None or not self.mandatory else ...,
                description=self.description,
                alias=self.get_qualified_name(),
            ),  # type: ignore[arg-type]
            cls=self.to_pydantic_model(),
        )

    def name(self) -> str:
        return self.make_unique_name(suffix="Leaf")

    def get_base_class(self) -> type | Node | Enum:
        base = TypeResolver.resolve_statement(self.raw_statement)
        return base

    def to_pydantic_model(self) -> type[BaseModel | RootModel]:
        """Generates the output class representing this node."""
        fields: Dict[str, Any] = self._children_to_fields()
        base: Any = self.get_base_class()

        if isinstance(base, Node):
            base = base._output_model.cls

        output_model: type[BaseModel] | type[RootModel]
        if base is not None:
            output_model = create_model(
                self.name(),
                __base__=(RootModel[base],),  # type: ignore[misc]
            )
        else:
            output_model = create_model(self.name(), __base__=(BaseModel,), **fields)
        output_model.__doc__ = self.description or ""
        return output_model


@NodeFactory.register_statement_class(["case"])
class CaseNode(Node):
    def __init__(self, stm: Statement) -> None:
        logger.debug(f"Parsing {__class__}")
        assert isinstance(stm, Statement)
        super().__init__(stm)

        self._output_model = GeneratedClass(
            class_name=self.name(),
            field_info=Field(..., alias=self.get_qualified_name()),
            cls=self.to_pydantic_model(),
        )

    def name(self) -> str:
        return self.make_unique_name(suffix="Case")

    def to_pydantic_model(self) -> type[BaseModel]:
        """Generates the output class representing this node."""
        fields: Dict[str, Any] = self._children_to_fields()
        output_model: type[BaseModel] = create_model(
            self.name(), __base__=(BaseModel,), **fields
        )
        output_model.__doc__ = self.description or ""
        return output_model


@NodeFactory.register_statement_class(["choice"])
class ChoiceNode(Node):
    def __init__(self, stm: ChoiceStatement) -> None:
        logger.debug(f"Parsing {__class__}")
        assert isinstance(stm, ChoiceStatement)
        super().__init__(stm)

        self._output_model = GeneratedClass(
            class_name=self.name(),
            field_info=Field(  # type: ignore[arg-type]
                ... if self.mandatory else None, alias=self.get_qualified_name()
            ),
            cls=self.to_pydantic_model(),
        )

    def name(self) -> str:
        return self.make_unique_name(suffix="Choice")

    def to_pydantic_model(self) -> type[BaseModel]:
        """Generates the output class representing this node."""
        fields: Dict[str, Any] = self._children_to_fields()
        fields.pop("namespace")
        fields.pop("prefix")
        bases: tuple = tuple(x[0] for x in fields.values())
        output_model: type[BaseModel] = Union[bases]  # type: ignore
        return output_model


@NodeFactory.register_statement_class(["container"])
class ContainerNode(Node):
    def __init__(self, stm: ContainerStatement) -> None:
        logger.debug(f"Parsing {__class__.__name__}")
        assert isinstance(stm, ContainerStatement)
        super().__init__(stm)

        self._output_model = GeneratedClass(
            class_name=self.name(),
            field_info=Field(  # type: ignore[arg-type]
                ... if self.mandatory else None,
                alias=self.get_qualified_name(),
            ),
            cls=self.to_pydantic_model(),
        )

    def name(self) -> str:
        return self.make_unique_name(suffix="Container")


@NodeFactory.register_statement_class(["list"])
class ListNode(Node):
    def __init__(self, stm: ListStatement) -> None:
        logger.debug(f"Parsing {__class__}")
        assert isinstance(stm, ListStatement)
        super().__init__(stm)

        keys: List[str] = self.__extract_keys(stm)
        if keys:
            for ch in self.children:
                if ch.arg in keys and isinstance(
                    ch._output_model.field_info, FieldInfo
                ):
                    ch.mandatory = True
                    new_field_info = Field(
                        ...,
                        alias=ch._output_model.field_info.alias,
                        description=ch._output_model.field_info.description,
                    )
                    ch._output_model.field_info = new_field_info

        output_class = self.to_pydantic_model()
        self._output_model = GeneratedClass(
            class_name=self.name(),
            cls=output_class,
            field_annotation=List[output_class],  # type: ignore
            field_info=Field(  # type: ignore[arg-type]
                ... if self.mandatory else [], alias=self.get_qualified_name()
            ),
        )

    @staticmethod
    def __extract_keys(stm: Statement) -> List[str]:
        """Returns the values of the key field."""
        key_stms = stm.search_one("key")
        return key_stms.arg.split() if key_stms is not None else []

    def name(self) -> str:
        return self.make_unique_name(suffix="ListEntry")

    def to_pydantic_model(self) -> type[BaseModel]:
        """Generates the output class representing this node."""
        fields: Dict[str, Any] = self._children_to_fields()
        output_model: type[BaseModel] = create_model(
            self.name(), __base__=(BaseModel,), **fields
        )
        output_model.__doc__ = self.description or ""
        return output_model


@NodeFactory.register_statement_class(["leaf-list"])
class LeafListNode(Node):
    def __init__(self, stm: LeafLeaflistStatement) -> None:
        logger.debug(f"Parsing {__class__}")
        super().__init__(stm)

        if self.default is None:
            default = ... if self.mandatory else []
        else:
            default = self.default

        output_class = self.to_pydantic_model()
        self._output_model = GeneratedClass(
            class_name=self.name(),
            cls=output_class,
            field_annotation=List[output_class],  # type: ignore
            field_info=Field(
                default,
                description=self.description,
                alias=self.get_qualified_name(),
            ),  # type: ignore[arg-type]
        )

    def name(self) -> str:
        return self.make_unique_name(suffix="LeafList")

    def get_base_class(self) -> type | Node | Enum:
        base = TypeResolver.resolve_statement(self.raw_statement)
        return base

    def to_pydantic_model(self) -> type[BaseModel | RootModel]:
        """Generates the output class representing this node."""
        fields: Dict[str, Any] = self._children_to_fields()
        base: Any = self.get_base_class()

        if isinstance(base, Node):
            base = base._output_model.cls

        output_model: type[BaseModel] | type[RootModel]
        if base is not None:
            output_model = create_model(
                self.name(),
                __base__=(RootModel[base],),  # type: ignore[misc]
            )
        else:
            output_model = create_model(self.name(), __base__=(BaseModel,), **fields)
        output_model.__doc__ = self.description or ""
        return output_model


@NodeFactory.register_statement_class(["module"])
class ModuleNode(Node):
    def __init__(self, stm: ModSubmodStatement) -> None:
        logger.debug(f"Parsing {__class__}")
        assert isinstance(stm, ModSubmodStatement)
        super().__init__(stm)

        output_class = self.to_pydantic_model()
        self._output_model = GeneratedClass(
            class_name=self.name(),
            cls=output_class,
            field_annotation=output_class,
            field_info=Field(...),
        )

    def to_pydantic_model(self) -> type[BaseModel] | type[RootModel]:
        return super().to_pydantic_model()

    def name(self) -> str:
        return self.make_unique_name(suffix="Module")


class ModelRoot:
    def __init__(self, stm: type[Statement]):
        self.root_node: Node | None = NodeFactory.generate(stm)

    def to_pydantic_model(self) -> type[BaseModel] | None:
        fields: Dict
        if isinstance(self.root_node, ModuleNode):
            # Take only children, as
            fields = self.root_node._children_to_fields()
        elif isinstance(self.root_node, Node):
            fields = {
                "namespace": (
                    str,
                    FieldInfo(
                        default=self.root_node.namespace,
                        json_schema_extra={"x-is-classvar": True},
                    ),
                ),
                "prefix": (
                    str,
                    FieldInfo(
                        default=self.root_node.prefix,
                        json_schema_extra={"x-is-classvar": True},
                    ),
                ),
                self.root_node.arg: self.root_node.get_output_class().to_field(),
            }
        output_model: type[BaseModel] = create_model(
            "Model", __base__=(BaseModel,), **fields
        )
        output_model.__doc__ = """
Initialize an instance of this class and serialize it to JSON; this results in a RESTCONF payload.

## Tips
Initialization:
- all values have to be set via keyword arguments
- if a class contains only a `root` field, it can be initialized as follows:
    - `member=MyNode(root=<value>)`
    - `member=<value>`

Serialziation:
- `exclude_defaults=True` omits fields set to their default value (recommended)
- `by_alias=True` ensures qualified names are used (necessary)
"""
        return output_model if fields else None



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple, Type

from datamodel_code_generator.reference import FieldNameResolver
from pyang.statements import (
    Statement,
)
from pyang.types import Decimal64Value
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, create_model
from pydantic import RootModel as PydanticRootModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined, PydanticUndefinedType

from ..utility.yang_sources_tracker import YANGSourcesTracker

if TYPE_CHECKING:
    __class__: Type

logger = logging.getLogger("pydantify")


def custom_json_schema_extra(schema: dict[str, Any]) -> None:  # type: ignore
    # Remove "title" property to avoid redundant annotation
    for prop in schema.get("properties", {}).values():
        prop.pop("title", None)
    if not schema.get("type", None) == "object":
        schema.pop("title", None)


model_config = ConfigDict(
    regex_engine="python-re",
    json_schema_extra=custom_json_schema_extra,
)


class BaseModel(PydanticBaseModel):
    model_config = model_config


class RootModel(PydanticRootModel):
    model_config = model_config


@dataclass
class GeneratedClass:
    """Holds information about a dynamically generated output class."""

    class_name: str | PydanticUndefinedType = PydanticUndefined
    """Output class name"""
    cls: Type[BaseModel] | Type[RootModel] | PydanticUndefinedType = PydanticUndefined
    """Ouput model class"""
    field_info: FieldInfo | PydanticUndefinedType = PydanticUndefined
    """Field info to add to field annotation"""
    field_annotation: Type | None = None
    """Annotated type when used """

    def assert_is_valid(self):
        for prop in self.__dataclass_fields__.keys():
            value = getattr(self, prop, PydanticUndefined)
            if value == PydanticUndefined:
                raise Exception(
                    f'Member "{prop}" of class "{__class__.__name__}" is PydanticUndefined.\n{self}'
                )

    def to_field(self) -> Tuple[Type[BaseModel] | Type, FieldInfo]:
        # Exception if any field are PydanticUndefined
        self.assert_is_valid()
        return (
            self.field_annotation if self.field_annotation else self.cls,
            self.field_info,
        )  # type: ignore


class Node(ABC):
    _name_count: Dict[str, int] = (
        dict()
    )  # keeps track of the number of models with the same name
    alias_mapping: Dict[str, str] = dict()
    data_type: Literal["config", "state"] | None
    strip_namespace: bool

    def __init__(self, stm: Statement):
        self.config: bool = __class__.__extract_config(stm)
        self.children: List[Node] = __class__.extract_statement_list(stm, "i_children")

        if stm.i_module:
            self.namespace = stm.i_module.search_one("namespace").arg
            self.prefix = stm.i_module.search_one("prefix").arg
        else:
            self.namespace = stm.search_one("namespace").arg
            self.prefix = stm.search_one("prefix").arg

        self.mandatory: bool = stm.search_one("mandatory", "true") or any(
            (ch for ch in self.children if ch.mandatory is True)
        )
        self.arg: str = stm.arg
        self.keyword: str = stm.keyword
        self.raw_statement: Statement = stm
        self.substmts: List[Statement] = stm.substmts
        self.comments: str | None = __class__.__extract_comments(stm)
        self.description: str | None = __class__.__extract_description(stm)

        def convert_default_values(stm_default: Any) -> Any:
            if isinstance(stm_default, Decimal64Value):
                return stm_default.value
            elif (
                isinstance(stm_default, Statement) and stm_default.keyword == "identity"
            ):
                return stm_default.arg
            else:
                return stm_default

        default = getattr(self.raw_statement, "i_default", PydanticUndefined)
        if isinstance(default, List):
            self.default = [convert_default_values(x) for x in default]
        else:
            self.default = convert_default_values(default)

        self._name: Optional[str] = None

        self._output_model: GeneratedClass = GeneratedClass()
        YANGSourcesTracker.track_from_pos(stm.pos)

    def get_output_class(self) -> GeneratedClass:
        return self._output_model

    def get_qualified_name(self) -> str:
        qualified_name = (
            f"{self.arg}"
            if self.strip_namespace
            else f"{self.raw_statement.i_module.arg}:{self.arg}"
        )
        self.alias_mapping[qualified_name] = FieldNameResolver(
            snake_case_field=True
        ).get_valid_name(self.arg)
        return qualified_name

    @abstractmethod
    def name(self) -> str:
        pass

    def make_unique_name(self, suffix: str):
        if self._name is None:
            self._name = Node.ensure_unique_name(f"{self.arg.capitalize()}{suffix}")
        return self._name

    @classmethod
    def ensure_unique_name(cls, name: str) -> str:
        count: int = Node._name_count.setdefault(name, 0)
        ret: str = name
        if count > 0:
            ret = f"{name}{count + 1}"
        Node._name_count[name] += 1
        return ret

    def get_base_class(self) -> type | Node | Enum:
        """Returns the class the output class should be derived from. Defaults to BaseModel."""
        return BaseModel

    @staticmethod
    def __extract_comments(stm: Statement) -> str | None:
        """Gathers and returns all comments located in the node's root, if present."""
        comments = stm.search("_comment")
        if comments:
            return " ".join((c.arg.lstrip("/ \n") for c in comments))
        return None

    @staticmethod
    def __extract_description(stm: Statement) -> str | None:
        """Returns the content of the "description" field, if present."""
        description = stm.search_one("description")
        return description.arg if description is not None else None

    @staticmethod
    def __extract_config(stm: Statement) -> bool:
        """Returns the content of the "config" field, if present."""
        config = getattr(stm, "i_config", None)
        return True if config is None else config

    def to_pydantic_model(self) -> Type[BaseModel] | Type[RootModel]:
        """Generates the output class representing this node."""
        fields: Dict[str, Any] = self._children_to_fields()
        base = self.get_base_class()

        if isinstance(base, Node) or isinstance(base, Enum):
            raise Exception(f"Base model needs to be a class. Got {base}")

        output_model: Type[BaseModel] = create_model(
            self.name(), __base__=(base,), **fields
        )
        output_model.__doc__ = self.description or ""
        return output_model

    def _children_to_fields(self) -> Dict[str, Tuple[type, FieldInfo]]:
        ret: Dict[str, Tuple[type, FieldInfo]] = dict()
        ret["namespace"] = (
            str,
            FieldInfo(
                default=self.namespace, json_schema_extra={"x-is-classvar": True}
            ),
        )
        ret["prefix"] = (
            str,
            FieldInfo(default=self.prefix, json_schema_extra={"x-is-classvar": True}),
        )
        for ch in self.children:
            if (
                self.data_type == "config"
                and ch.config is True
                or self.data_type == "state"
                and ch.config is False
                or self.data_type is None
                or ch.mandatory is True
            ):
                ret[ch.arg] = ch._output_model.to_field()

        if len(ret) > 2 and self.data_type == "state":
            self.config = False

        return ret

    @staticmethod
    def extract_statement_list(statement: Statement, attr_name: str) -> List[Node]:
        from .nodefactory import NodeFactory

        rv = getattr(statement, attr_name, [])
        return [ch for ch in map(NodeFactory.generate, rv) if ch is not None]



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Type

from pyang.statements import Statement
from typing_extensions import Self

from pydantify.exceptions import NotImplementedException

from . import Node, TypeResolver

logger = logging.getLogger("pydantify")


class NodeFactory:
    # src: statements.data_definition_keywords
    @dataclass
    class ClassMapping:
        maps_to: Callable[..., Node]

        def __call__(self, *args: Any, **kwds: Any) -> Any:
            return self.maps_to(*args, **kwds)

    _implemented_mappings: Dict[str, ClassMapping] = {}
    _ignored_types: List[str] = [
        "rpc",
        "notification",
    ]  #  TODO: potential extension to project?

    @classmethod
    def register_statement_class(cls: Type[Self], keywords: List[str]):
        for keyword in keywords:
            assert keyword not in cls._implemented_mappings.keys()

        def _register(type: type):
            for keyword in keywords:
                cls._implemented_mappings[keyword] = cls.ClassMapping(maps_to=type)
            return type

        return _register

    @classmethod
    def generate(cls: Type[Self], stm: Statement) -> Node | None:
        assert isinstance(stm, Statement)
        known_model = TypeResolver.get_model_if_known(stm)
        if known_model is not None:
            return known_model
        else:
            if stm.keyword in cls._ignored_types:
                return None
            if stm.keyword not in cls._implemented_mappings.keys():
                raise NotImplementedException(
                    f'"{stm.keyword}" has not yet been implemented as a type.'
                )
            mapping = cls._implemented_mappings[stm.keyword]
            node = mapping.maps_to(stm)
            TypeResolver.register(stm, node)
            return node
